import uuid
from abc import ABC, abstractmethod
from contextlib import suppress
from dataclasses import dataclass
from functools import lru_cache, cached_property
from inspect import signature, Parameter, isabstract
from pydoc import locate
from typing import (
    Deque, Dict, FrozenSet, List, Optional, Sequence, Set, Tuple, Union, get_args, Callable
)
from typing import Type, get_origin

from pydantic import BaseModel, BaseConfig
from pydantic.fields import ModelField
from river.base import Classifier, DriftDetector, Estimator
from river.datasets.base import Dataset
from river.metrics.base import Metric
from river.optim.base import Initializer, Scheduler, Optimizer

from cacp.gui.external.shared.type import get_all_non_abstract_subclasses, is_primitive_type, to_id

NONE_TYPE = type(None)
SEARCHABLE = [
    Dataset,
    Classifier,
    DriftDetector,
    Optimizer,
    Metric,
    Estimator,
    Initializer,
    Scheduler
]


@dataclass
class ToPydanticMapper(ABC):
    parameter: Parameter

    @abstractmethod
    def can_map(self) -> bool:
        pass

    @abstractmethod
    def map(self) -> ModelField:
        pass

    @cached_property
    def annotation(self):
        args = get_args(self.parameter.annotation)
        if len(args) == 2:
            if args[0] is NONE_TYPE:  # handle type.Optional
                return args[1]
            elif args[1] is NONE_TYPE:  # handle type.Optional
                return args[0]
        return self.parameter.annotation

    @cached_property
    def default(self):
        return self.parameter.default if is_primitive_type(
            self.annotation) and self.parameter.default != self.parameter.empty else None


@dataclass
class SearchableToPydanticMapper(ToPydanticMapper):
    searchable: Optional[Type] = None

    def can_map(self) -> bool:
        annotation = self.annotation
        with suppress(TypeError):
            for searchable in SEARCHABLE:
                if issubclass(annotation, searchable):
                    self.searchable = searchable
                    return True

        return False

    def map(self) -> None:
        # disabled for now, deep level of setup is not required, and is causing recursion
        return None

    @staticmethod
    @lru_cache
    def _searchable_type_to_pseudo_pydantic_model(searchable, annotation):
        model = new_pseudo_pydantic_model(f"{searchable.__name__}__{uuid.uuid4().hex}")
        subclasses = get_all_non_abstract_subclasses(annotation)
        if not subclasses:
            subclasses.append(annotation)
        pseudo_subclasses = [
            new_pseudo_pydantic_model(f"{to_id(x)}__{uuid.uuid4().hex}") for x in
            subclasses
        ]
        p_annotation = Union[tuple(pseudo_subclasses)]
        model.__fields__["option"] = ModelField.infer(
            name="option",
            value=None,
            annotation=p_annotation,
            class_validators=None,
            config=BaseConfig,
        )
        return model


@dataclass
class IterableToPydanticMapper(ToPydanticMapper):
    ITERABLE_TYPES = [
        Deque,
        Dict,
        dict,
        FrozenSet,
        List,
        list,
        Optional,  # Union[x,None]
        Sequence,
        Set,
        set,
        Tuple,
        Union,
    ]

    def can_map(self) -> bool:
        annotation = self.annotation
        origin = get_origin(annotation)
        return origin in self.ITERABLE_TYPES or annotation in self.ITERABLE_TYPES

    def map(self) -> ModelField:
        args = get_args(self.annotation)
        if not args:
            return None
        if len(args) == 1 and args[0] is NONE_TYPE:  # handle type.Optional
            annotation = type_to_pseudo_pydantic_model(args[1], self.parameter.name)
        elif len(args) == 2 and args[1] is NONE_TYPE:  # handle type.Optional
            annotation = type_to_pseudo_pydantic_model(args[0], self.parameter.name)
        else:  # handle other types
            pseudo_values = [type_to_pseudo_pydantic_model(x, self.parameter.name) for x in args]
            annotation = Union[tuple(pseudo_values)]

        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=annotation,
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class AnyAbstractToPydanticMapper(ToPydanticMapper):

    def can_map(self) -> bool:
        return isabstract(self.annotation)

    def map(self) -> ModelField:
        subclasses = get_all_non_abstract_subclasses(self.annotation)
        pseudo_subclasses = [type_to_pseudo_pydantic_model(x, self.parameter.name) for x in subclasses]
        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=Union[tuple(pseudo_subclasses)],
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class AnyToPydanticMapper(ToPydanticMapper):

    def can_map(self) -> bool:
        if "typing.Callable" in str(self.annotation):
            return False
        return True

    def map(self) -> ModelField:
        annotation = type_to_pseudo_pydantic_model(self.annotation, self.parameter.name)
        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=annotation,
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class DocToPydanticMapper:

    @classmethod
    def map(cls, class_type, parameter) -> ModelField:
        doc_string = class_type.__doc__
        if doc_string:
            parameter_in_doc_string = f"{parameter.name} :"
            for l in doc_string.split("\n"):
                l = l.strip()
                if l.startswith(parameter_in_doc_string):
                    possible_primitive = l.replace(parameter_in_doc_string, "").strip().split(",")[0]
                    if possible_primitive in ["bool", "int", "float", "str"]:
                        return ModelField.infer(
                            name=parameter.name,
                            value=None,
                            annotation=locate(possible_primitive),
                            class_validators=None,
                            config=BaseConfig,
                        )


MAPPER_CLASSES = [
    SearchableToPydanticMapper,
    IterableToPydanticMapper,
    AnyAbstractToPydanticMapper,
    AnyToPydanticMapper
]


def new_pseudo_pydantic_model(name: str):
    class Model(BaseModel):
        pass

    Model.__name__ = name
    return Model


@lru_cache
def function_to_pseudo_pydantic_model(function: Callable):
    model = new_pseudo_pydantic_model(function.__name__)
    sig = signature(function)
    parameters: List[Parameter] = list(sig.parameters.values())
    for parameter in parameters:
        if parameter.annotation == parameter.empty and parameter.default != parameter.empty:
            potential_annotation_type = type(parameter.default)
            if is_primitive_type(potential_annotation_type):
                parameter._annotation = potential_annotation_type
        if parameter.annotation != parameter.empty:
            for mapper_class in MAPPER_CLASSES:
                mapper = mapper_class(parameter)
                if mapper.can_map():
                    field = mapper.map()
                    if field:
                        model.__fields__[parameter.name] = field
                    break
        else:
            field = DocToPydanticMapper.map(function, parameter)
            if field:
                model.__fields__[parameter.name] = field
    return model


@lru_cache
def type_to_pseudo_pydantic_model(t: Type, alternative_name: str = None):
    if is_primitive_type(t):
        return t

    model = new_pseudo_pydantic_model(t.__name__ if hasattr(t, "__name__") else alternative_name)
    sig = signature(t.__init__)
    init_parameters: List[Parameter] = list(sig.parameters.values())[1:]
    for parameter in init_parameters:
        if parameter.annotation == parameter.empty and parameter.default != parameter.empty:
            potential_annotation_type = type(parameter.default)
            if is_primitive_type(potential_annotation_type):
                parameter._annotation = potential_annotation_type
        if parameter.annotation != parameter.empty:
            for mapper_class in MAPPER_CLASSES:
                mapper = mapper_class(parameter)
                if mapper.can_map():
                    field = mapper.map()
                    if field:
                        model.__fields__[parameter.name] = field
                    break
        else:
            field = DocToPydanticMapper.map(t, parameter)
            if field:
                model.__fields__[parameter.name] = field

    return model



from typing import Type

import river
from river.metrics.base import ClassificationMetric

from cacp.gui.external.shared.model import ClassModel
from cacp.gui.external.shared.type import to_id


class RiverMetricModel(ClassModel):
    name: str
    docs_url: str

    @classmethod
    def base_class(cls):
        return ClassificationMetric

    @classmethod
    def from_class(cls, source_class: Type) -> "RiverClassifierModel":
        _id = to_id(source_class)
        docs_split = _id.split(".")
        docs_version = river.__version__
        docs_name = docs_split[3] if len(docs_split) > 3 else docs_split[2]
        return cls(
            id=_id,
            name=source_class.__name__,
            docs_url=f"https://riverml.xyz/{docs_version}/api/{docs_split[1].replace('_', '-')}/{docs_name}/"
        )


if __name__ == '__main__':
    me = RiverMetricModel.all()
    print(me)



import shutil
from datetime import datetime
from enum import Enum
from typing import TypedDict, List, Dict, Optional

from tinydb import TinyDB
from tinydb.table import Document

from cacp.gui.db import DB_PATH


class ExperimentType(str, Enum):
    BATCH = "BATCH"
    INCREMENTAL = "INCREMENTAL"


class ExperimentStatus(str, Enum):
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


class Experiment(TypedDict):
    id: int
    path: Optional[str]
    name: str
    datasets: List[Dict]
    classifiers: List[Dict]
    metrics: List[Dict]
    type: ExperimentType
    status: ExperimentStatus
    created_at: float


EXPERIMENTS_DB = TinyDB(DB_PATH / "experiments.json")
EXPERIMENTS_PATH = (DB_PATH / "experiments").resolve()


def _convert_from_document_to_experiment(experiment: Document) -> Experiment:
    experiment["id"] = experiment.doc_id
    experiment["number of datasets"] = len(experiment["datasets"])
    experiment["number of classifiers"] = len(experiment["classifiers"])
    experiment["created at"] = datetime.fromtimestamp(experiment["created_at"]).strftime("%Y-%m-%d %H:%M:%S")
    return experiment


def get_all_experiments() -> List[Experiment]:
    return [_convert_from_document_to_experiment(e) for e in EXPERIMENTS_DB.all()]


def get_experiment(experiment_id: int) -> Experiment:
    return _convert_from_document_to_experiment(EXPERIMENTS_DB.get(doc_id=experiment_id))


def add_experiment(
    name_value: str, type_value: ExperimentType, selected_datasets: List[Dict], selected_classifiers: List[Dict],
    selected_metrics: List[Dict]
) -> int:
    new_experiment: Experiment = dict()
    new_experiment["name"] = name_value
    new_experiment["type"] = type_value
    new_experiment["datasets"] = selected_datasets
    new_experiment["classifiers"] = selected_classifiers
    new_experiment["metrics"] = selected_metrics
    new_experiment["status"] = ExperimentStatus.RUNNING
    new_experiment["created_at"] = datetime.now().timestamp()
    experiment_id = EXPERIMENTS_DB.insert(new_experiment)
    EXPERIMENTS_DB.update({"path": str((EXPERIMENTS_PATH / str(experiment_id)).as_posix())}, doc_ids=[experiment_id])
    return experiment_id


def update_experiment_status(experiment_id: int, experiment_status: ExperimentStatus):
    EXPERIMENTS_DB.update({"status": experiment_status}, doc_ids=[experiment_id])


def delete_experiment(experiment_id: int):
    experiment = get_experiment(experiment_id)
    shutil.rmtree(experiment["path"], ignore_errors=True)
    EXPERIMENTS_DB.remove(doc_ids=[experiment_id])



from pathlib import Path

import dash
from dash import html, Output, Input, no_update, callback
from dash.dcc import Store

from cacp.gui.components.experiments.details.processor import ExperimentProcessor
from cacp.gui.components.shared.utils import csv_to_grid
from cacp.gui.db.experiments import get_experiment, Experiment

dash.register_page(__name__, path_template="/experiment/<experiment_id>")

EXPERIMENT_STORE_ID = "experiment_store"
EXPERIMENT_RESULTS_ID = "experiment_results"


def show_experiment(experiment: Experiment):
    experiment_path = Path(experiment["path"])
    return html.Div([
        html.H5("Datasets", className="mt-4"),
        csv_to_grid(experiment_path / "info" / "datasets.csv", skip_first_column=True),
        html.H5("Classifiers", className="mt-4"),
        csv_to_grid(experiment_path / "info" / "classifiers.csv", skip_first_column=True),
        html.H5("Comparison results", className="mt-4"),
        csv_to_grid(experiment_path / "comparison_result.csv", skip_first_column=True),
        html.H5("Comparison results winner", className="mt-4"),
        csv_to_grid(experiment_path / "winner" / "comparison.csv", skip_first_column=True),
        html.H5("Comparison times", className="mt-4"),
        csv_to_grid(experiment_path / "time" / "comparison.csv"),
    ])


EXPERIMENT_PROCESSOR = ExperimentProcessor(EXPERIMENT_STORE_ID)


def layout(experiment_id=None):
    experiment = get_experiment(experiment_id) if experiment_id else None
    view = html.Div([
        EXPERIMENT_PROCESSOR,
        html.Div(show_experiment(experiment) if experiment else [], id=EXPERIMENT_RESULTS_ID),
        Store(id=EXPERIMENT_STORE_ID, data=experiment),
    ])
    return view


@callback(
    Output(EXPERIMENT_RESULTS_ID, "children"),
    Input(EXPERIMENT_STORE_ID, "data")
)
def on_experiment_changed(experiment: Experiment):
    if experiment:
        return show_experiment(experiment)
    return no_update



import logging

import dash_bootstrap_components as dbc
import diskcache
import flask
from dash import Dash, html
from dash import DiskcacheManager
from dash import page_container
from dash.dcc import Location

from cacp.gui.components.shared.sidebar import sidebar_component
from cacp.gui.components.shared.utils import GLOBAL_LOCATION_ID
from cacp.gui.db import DB_PATH

CACHE_PATH = DB_PATH / ".cache"
CACHE_PATH.mkdir(exist_ok=True, parents=True)
BACKGROUND_CALLBACK_MANAGER = DiskcacheManager(diskcache.Cache(CACHE_PATH))


def start(debug: bool = False, host="127.0.0.1", port=8050):
    log_level = logging.INFO if debug else logging.ERROR
    flask_log = logging.getLogger("werkzeug")
    flask_log.setLevel(log_level)
    flask.cli.show_server_banner = lambda *_args, **_kwargs: None

    app = Dash(
        __name__,
        external_stylesheets=[dbc.themes.COSMO, dbc.icons.BOOTSTRAP],
        use_pages=True,
        background_callback_manager=BACKGROUND_CALLBACK_MANAGER
    )
    app.layout = html.Div([
        Location(id=GLOBAL_LOCATION_ID),
        sidebar_component(),
        html.Div(id="page-content", style={
            "margin-left": "18rem",
            "margin-right": "2rem",
            "padding": "2rem 1rem",
        }, children=[page_container]),

    ])
    app.logger.setLevel(log_level)

    print(f"CACP stared on http://{host}:{port}/")
    app.run_server(debug=debug, host=host, port=port)


if __name__ == "__main__":
    start(debug=True, host="127.0.0.1", port=8050)



from inspect import isabstract
from typing import Type, Callable, Union


def to_id(value: Union[Type, Callable]):
    return value.__module__ + "." + value.__name__


def get_all_subclasses(cls: Type):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def get_all_non_abstract_subclasses(cls):
    all_subclasses = get_all_subclasses(cls)
    return [s for s in all_subclasses if not isabstract(s)]


PRIMITIVE_TYPES = (bool, int, float, complex, str)


def is_primitive_type(thing):
    return thing in PRIMITIVE_TYPES
